/*
Copyright (c) 2001 Dan Vernon

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version. This program is
distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. You should have received a copy of the GNU
General Public License along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// sioProtocol.cpp: implementation of the CSIOProtocol class.
//
//////////////////////////////////////////////////////////////////////

#include "sioProtocol.h"
#include "sharePoint.h"
#include "time.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


CSIOProtocol::CSIOProtocol()
{
	timer = new CuTimer();
	sioPort = new CSIOPort();

	for(int i=0;i<SIO_MAX_DISKS+1;i++)
		disk[i] = NULL;

	printer = new CDiskImage();

	//init bad sector maps
	memset(badSectorMapSD,0,sizeof(badSectorMapSD));
	badSectorMapSD[0]=0xff;
	badSectorMapSD[1]=0xff;
	badSectorMapSD[SIO_SD_SECTOR_SIZE]=checkSum(badSectorMapSD,SIO_SD_SECTOR_SIZE);

	memset(badSectorMapDD,0,sizeof(badSectorMapDD));
	badSectorMapDD[0]=0xff;
	badSectorMapDD[1]=0xff;
	badSectorMapDD[SIO_DD_SECTOR_SIZE]=checkSum(badSectorMapDD,SIO_DD_SECTOR_SIZE);

	diskMutex = CreateMutex(NULL, false, NULL);
}

CSIOProtocol::~CSIOProtocol()
{
	WaitForSingleObject(diskMutex, INFINITE);

	for(int i=1;i<=SIO_MAX_DISKS;i++)
	{
		if(disk[i])
			release(i);
	}
	delete timer;

	delete sioPort;

	ReleaseMutex(diskMutex);

	CloseHandle(diskMutex);
}

bool CSIOProtocol::mount(LPSTR path, BYTE diskid, bool readOnly)
{
	bool success = false;
	DWORD attr=0;
	LPSTR t=NULL;

	WaitForSingleObject(diskMutex, INFINITE);

	if (diskid >= 1 && diskid <= SIO_MAX_DISKS)
	{
		attr=GetFileAttributes(path);

		if (attr!=-1) //file not found
		{
			if (disk[diskid])
				release(diskid);

			if (attr & FILE_ATTRIBUTE_DIRECTORY)
				disk[diskid]=new CSharePoint();
			else
				disk[diskid]=new CDiskImage();
				
			success = disk[diskid]->load(path,readOnly);

			CDiskImage* di = disk[diskid];
			if (success) 
			{
				//t=strrchr(path,'\\');
   				innerShell_status("%s mounted as D%d:%s",path,diskid,di->description());
			}
			else
			{
				if (disk[diskid])
					release(diskid);
			}
		}
	}

	ReleaseMutex(diskMutex);

	return success;
}

bool CSIOProtocol::release(BYTE diskid)
{
	bool success = false;

	WaitForSingleObject(diskMutex, INFINITE);

	if (diskid >= 1 && diskid <= SIO_MAX_DISKS && disk[diskid])
	{
		delete disk[diskid];
		disk[diskid] = NULL;
		success = true;
	   	innerShell_status("D%d: released",diskid);
	}

	ReleaseMutex(diskMutex);

	return success;
}


void CSIOProtocol::listen()
{
	BYTE cmd[SIO_COMMAND_SIZE];

	//try
	//{
		while(!g_state->isCancelSet()
			  && !g_state->isErrorSet())
		{
			g_state->setHappyRWMode(false); // this flag is set on happy-read, happy-write
											// (disables high-speed toggling on framing error)

			if (sioPort->readBytes(cmd,SIO_COMMAND_SIZE))
				if (isValidCommand(cmd))
					handleCommand(cmd);
		}
	//}
	//catch(...){} //throw is only used for flow control
}

bool CSIOProtocol::isValidCommand(BYTE* cmd)
{
	LPCSTR src = "CSIOProtocol::isValidCommand";
	if (cmd[SIO_COMMAND_SIZE-1]!=checkSum(cmd,SIO_COMMAND_SIZE-1))
	{
		char msg[255];
		sprintf(msg,"%02x %02x %02x %02x %02x -- BAD SUM in command frame",cmd[0],cmd[1],cmd[2],cmd[3],cmd[4]);
		innerShell_debug(msg);
		sioPort->resetPort();
		return false;
	}
	else
	{
		return true;
	}
}


bool CSIOProtocol::setConfiguration(CDiskImage* di)
{
	LPCSTR src = "CSIOProtocol::writeSector";
	DEVICE_CONFIG_BLOCK dcb;
	bool res=true;

	res=sioPort->readBytes((BYTE*)&dcb,sizeof(dcb));

	if(res)
	{
		if (dcb.checkSum != checkSum((BYTE*)&dcb,sizeof(dcb)-1))
		{
			innerShell_debug("BAD SUM - data frame");
			sioPort->resetPort();
			res=false;
		}
		else
		{
			res=di->setDeviceConfig(&dcb);
		}
	}
	return res;
}

bool CSIOProtocol::readConfiguration(CDiskImage* di)
{
	LPCSTR src = "CSIOProtocol::readSector";
	bool res=true;
	res&=sioPort->writeBytes((BYTE*)(di->getDeviceConfig()),
		                        sizeof(DEVICE_CONFIG_BLOCK));
	return res;
}


bool CSIOProtocol::readSector(CDiskImage* di, WORD sector)
{
	BYTE chunk[SIO_MAX_SECTOR_SIZE];
	LPCSTR src = "CSIOProtocol::readSector";
	DWORD size;
	bool res=true;

	size=di->readSector(chunk,sector);
	chunk[size]=checkSum(chunk,size);
	res&=sioPort->writeBytes(chunk,size+1);

	return res;
}


bool CSIOProtocol::writeSector(CDiskImage* di, WORD sector)
{
	LPCSTR src = "CSIOProtocol::writeSector";
	BYTE chunk[SIO_MAX_SECTOR_SIZE];
	DWORD size;
	bool res=true;
	size=di->getSectorSize(sector);

	res=sioPort->readBytes(chunk,size+1);

	if(res)
	{
		if (chunk[size] != checkSum(chunk,size))
		{
			innerShell_debug("BAD SUM - data frame");
			sioPort->resetPort();
			res=false;
		}
		else
		{
			res=(size==di->writeSector(chunk,sector));
		}
	}
	return res;
}

bool CSIOProtocol::printData()
{
	LPCSTR src = "CSIOProtocol::writeSector";
	BYTE chunk[SIO_MAX_SECTOR_SIZE];
	DWORD size;
	bool res=true;
	size=40; //SIO_SD_SECTOR_SIZE;

	res=sioPort->readBytes(chunk,size+1);

	if(res)
	{
		if (chunk[size] != checkSum(chunk,size))
		{
			innerShell_debug("BAD SUM - data frame");
			sioPort->resetPort();
			res=false;
		}
		else
		{
			innerShell_debug("%d bytes received for printing",size);
		}
	}
	else
	{
			innerShell_debug("failed to read data from printer");
	}
	return res;
}

BYTE CSIOProtocol::checkSum(BYTE* chunk,DWORD size)
{
	DWORD chkSum=0;
	for(DWORD i=0; i<size; i++ )
		chkSum=((chkSum+chunk[i])>>8)
		       +((chkSum+chunk[i])&0xff);
	return (BYTE)chkSum;
}

LPCSTR CSIOProtocol::xlateCommand(BYTE b,BYTE aux0,BYTE aux1)
{
	switch(b)
	{
	case 'R':return "read";    break;
	case 'W':return "write";   break;
	case 'S':return "status";  break;
	case 'P':return "put";     break;
	case '!':return "format";  break;
	case '"':return "format-ED";  break;
	case 'N':return "get-config"; break;
	case 'O':return "set-config"; break;
	case '?':return "speed-up"; break;
	case 'r':return "happy-read"; break;
	case 'w':return "happy-write"; break;
	case 'T':return "read-memory"; break;
	case 'U':return "motor-on"; break;
	case 'Q':return "quit"; break;
	case 'H':
		switch(aux0)
		{
			case 0x01:return "happy-motor-delay"; break;
			case 0x02:return "happy-set-driveid"; break;
			case 0x03:return "happy-clear-settings"; break;
			case 0x18:return ((aux1==10)?"happy-protect-on":"happy-protect-off"); break;
			case 0x20:return "happy-use-PUT-cmd"; break;
			case 0x60:return "happy-disable-ram"; break;
			case 0x00:return "happy-enable-ram"; break;
			case 0xe0:return "happy-set-unhappy"; break;
			default:return "happy-config"; break;
		}
		break;
	case 0x93:return "get-time"; break;
	default: return "unknown?"; break;
	}
}

bool CSIOProtocol::isKnownCommand(BYTE b)
{
	switch(b)
	{
	case 'R':return true; break;
	case 'W':return true; break;
	case 'S':return true; break;
	case 'P':return true; break;
	case '!':return true; break;
	case '"':return true; break;
	case 'N':return true; break;
	case 'O':return true; break;
	case '?':return true; break;
	case 'r':return true; break;
	case 'w':return true; break;
	case 'H':return true; break;
	case 'T':return true; break;
	case 'U':return true; break;
	case 'Q':return true; break;
	case 0x93:return true; break;
	default: return false; break;
	}
}

void CSIOProtocol::handleCommand(BYTE* cmd)
{
	WORD sec;
	BYTE id=0;
	BYTE aux0, aux1;


	id = cmd[0]-'0';
	aux0 = cmd[2];
	aux1 = cmd[3];
	sec = aux0 + 256*aux1;

	if((id>SIO_MAX_DISKS || id<1) && id!=0x10)
	{
		/*
		innerShell_debugSioCommand(cmd[0],cmd[1],cmd[2],cmd[3],cmd[4],
						       (BYTE)id,sec,xlateCommand(cmd[1],aux0,aux1),g_state->getCurrentSpeed());
		innerShell_debug("unknown device - %02x",id);
		*/
		return;
	}

	if(!isKnownCommand(cmd[1]))
	{
		innerShell_debugSioCommand(cmd[0],cmd[1],cmd[2],cmd[3],cmd[4],
						       (BYTE)id,sec,xlateCommand(cmd[1],aux0,aux1),g_state->getCurrentSpeed());
		return;
	}


	innerShell_sioCommand(cmd[0],cmd[1],cmd[2],cmd[3],cmd[4],
		              (BYTE)id,sec,xlateCommand(cmd[1],aux0,aux1),g_state->getCurrentSpeed());

	if (id==0x10) //print
	{
			switch(cmd[1])
			{
			case 'W': //write
			case 'P': //put
				timer->usleep(SIO_WAIT_CMD_ACK);
				if(!sioPort->writeAck('A')) break;
				if (printData())
				{
					timer->usleep(SIO_WAIT_DATA_ACK);
					if(!sioPort->writeAck('A')) break;
				}
				else
				{
					if (sioPort->ringInterruptFlag) break;
					timer->usleep(SIO_WAIT_DATA_ACK);
					if (!sioPort->writeAck('N')) break;
					sioPort->resetPort(); /*@@ test @@*/
					innerShell_debug("[NAK - data frame]");
					break;
				}
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->writeAck('C');
				break;

			case 'S': //device status
				timer->usleep(SIO_WAIT_CMD_ACK);
				if (!sioPort->writeAck('A')) break;
				timer->usleep(SIO_WAIT_COMPLETE);
				if (!sioPort->writeAck('C')) break;
				sioPort->writeBytes((BYTE*)(printer->getDeviceStatus()),sizeof(DEVICE_STATUS_BLOCK));
				break;

			default:
				timer->usleep(SIO_WAIT_CMD_ACK);
				sioPort->writeAck('N');
				break;
			}

	}
	else
	{
		WaitForSingleObject(diskMutex, INFINITE);
		
		if (!disk[id])
			;
			//{innerShell_status( "[no image mounted]" );}
		else
			switch(cmd[1])
			{

			case 'R': //read
				timer->usleep(SIO_WAIT_CMD_ACK);
				if (sec>disk[id]->getSectorCount())
				{
					if(!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - invalid sector]");
					break;
				}
				if(!sioPort->writeAck('A')) break;
				timer->usleep(SIO_WAIT_COMPLETE);
				if(!sioPort->writeAck('C')) break;
				readSector(disk[id],sec);
				break;

			case 'W': //write
			case 'P': //put
				if (disk[id]->isReadOnly())
				{
					if(!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - read-only image]");
					break;
				}
				if (sec>disk[id]->getSectorCount())
				{
					if(!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - invalid sector]");
					break;
				}
				timer->usleep(SIO_WAIT_CMD_ACK);
				if(!sioPort->writeAck('A')) break;
				if (writeSector(disk[id],sec))
				{
					timer->usleep(SIO_WAIT_DATA_ACK);
					if(!sioPort->writeAck('A')) break;
				}
				else
				{
					if (sioPort->ringInterruptFlag) break;
					timer->usleep(SIO_WAIT_DATA_ACK);
					if (!sioPort->writeAck('N')) break;
					sioPort->resetPort(); /*@@ test @@*/
					innerShell_debug("[NAK - data frame]");
					break;
				}
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->writeAck('C');
				break;

			case 'S': //device status
				timer->usleep(SIO_WAIT_CMD_ACK);
				if (!sioPort->writeAck('A')) break;
				timer->usleep(SIO_WAIT_COMPLETE);
				if (!sioPort->writeAck('C')) break;
				sioPort->writeBytes((BYTE*)(disk[id]->getDeviceStatus()),sizeof(DEVICE_STATUS_BLOCK));
				break;

			case '"': //format disk ED
				if(disk[id]->getSectorCount()==SIO_SSSD_SECTOR_NUM)
				{
					timer->usleep(SIO_WAIT_CMD_ACK);
					sioPort->writeAck('N');
					break;
				}
			case '!': //format disk SD/DD/??
				timer->usleep(SIO_WAIT_CMD_ACK);
				sioPort->writeAck('A');
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->writeAck('C');
				if(disk[id]->getSectorSize()==SIO_DD_SECTOR_SIZE)
					sioPort->writeBytes(badSectorMapDD,sizeof(badSectorMapDD));
				else
					sioPort->writeBytes(badSectorMapSD,sizeof(badSectorMapSD));
				break;

			case 'N': //read configuration
				timer->usleep(SIO_WAIT_CMD_ACK);
				if(!sioPort->writeAck('A')) break;
				timer->usleep(SIO_WAIT_COMPLETE);
				if(!sioPort->writeAck('C')) break;
				readConfiguration(disk[id]);
				break;

			case 'O': //set configuration
				timer->usleep(SIO_WAIT_CMD_ACK);
				if(!sioPort->writeAck('A')) break;
				if (setConfiguration(disk[id]))
				{
					timer->usleep(SIO_WAIT_DATA_ACK);
					if(!sioPort->writeAck('A')) break;
				}
				else
				{
					if (sioPort->ringInterruptFlag) break;
					timer->usleep(SIO_WAIT_DATA_ACK);
					if (!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - data frame]");
					break;
				}
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->writeAck('C');
				break;

			case 'r': //happy read
				sioPort->writeAck('N'); 
				break;

				g_state->setHappyRWMode(true);
				timer->usleep(SIO_WAIT_CMD_ACK);
				if (sec>disk[id]->getSectorCount())
				{
					if(!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - invalid sector]");
					break;
				}
				if(!sioPort->writeAck('A')) break;
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->setSpeed(CBR_38400);
				if(!sioPort->writeAck('C')) break;
				readSector(disk[id],sec);
				sioPort->setSpeed(CBR_19200);
				break;

			case 'w': //happy write
				sioPort->writeAck('N'); 
				break;

				g_state->setHappyRWMode(true);
				if (disk[id]->isReadOnly())
				{
					if(!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - read-only image]");
					break;
				}
				if (sec>disk[id]->getSectorCount())
				{
					if(!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - invalid sector]");
					break;
				}
				timer->usleep(SIO_WAIT_CMD_ACK);
				if(!sioPort->writeAck('A')) break;
				sioPort->setSpeed(CBR_38400);
				if (writeSector(disk[id],sec))
				{
					timer->usleep(SIO_WAIT_DATA_ACK);
					if(!sioPort->writeAck('A')) break;
					sioPort->setSpeed(CBR_19200);
				}
				else
				{
					if (sioPort->ringInterruptFlag) break;
					timer->usleep(SIO_WAIT_DATA_ACK);
					if (!sioPort->writeAck('N')) break;
					innerShell_debug("[NAK - data frame]");
					sioPort->setSpeed(CBR_19200);
					break;
				}
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->writeAck('C');
				break;

			case 'H': //happy config command

				if (sioPort->maxSpeed!=3)  // Happy 2X emulation is not working, so NAK it.
										// APE Warp driver at 3X works OK though.	
				{
					sioPort->writeAck('N'); 
					break;
				}

				timer->usleep(SIO_WAIT_CMD_ACK);
				sioPort->writeAck('A');
				disk[id]->setHappy(true);
				timer->usleep(SIO_WAIT_COMPLETE);
				sioPort->writeAck('C');
				break;

			case 0x93:
				//get current time (APE)
				BYTE apeTime[7];

				timer->usleep(SIO_WAIT_CMD_ACK);
				sioPort->writeAck('A');
				timer->usleep(SIO_WAIT_COMPLETE);
				if (!sioPort->writeAck('C')) break;

				struct tm *newtime;
				time_t long_time;

				time( &long_time );                /* Get time as long integer. */
				newtime = localtime( &long_time ); /* Convert to local time. */
				
				apeTime[0]=newtime->tm_mday;
				apeTime[1]=newtime->tm_mon+1;
				apeTime[2]=newtime->tm_year-(newtime->tm_year/100)*100;
				apeTime[3]=newtime->tm_hour;
				apeTime[4]=newtime->tm_min;
				apeTime[5]=newtime->tm_sec;
				apeTime[6]=checkSum(apeTime,6);
				sioPort->writeBytes(apeTime,sizeof(apeTime));
				break;

			case '?':
				if (sioPort->maxSpeed==1)
				{
					innerShell_debug("[request for high-speed SIO denied]");
					sioPort->writeAck('N');
				}
				else
				{

					BYTE speedIndex[2];
					timer->usleep(SIO_WAIT_CMD_ACK);
					sioPort->writeAck('A');
					timer->usleep(SIO_WAIT_COMPLETE);
					if (!sioPort->writeAck('C')) break;

					innerShell_debug((sioPort->maxSpeed==2)?"[jump to 2X]":"[jump to 3X]");

					speedIndex[0]=(sioPort->maxSpeed==2)?0x10:0x08;
					speedIndex[1]=speedIndex[0];  //i.e. checkSum(speedIndex,1);
					sioPort->writeBytes(speedIndex,sizeof(speedIndex));

					//if (!disk[id]->isHappy())
					//{
						//innerShell_debug("changing sioPort speed...");

						sioPort->setSpeed((sioPort->maxSpeed==2)?CBR_38400:CBR_57600);
						g_state->setCurrentSpeed(sioPort->maxSpeed);
						g_state->speedUpActive=true;
					/*}
					else
					{
						g_state->speedUpActive=false;
					}
					*/
				}
				break;

			default:
				timer->usleep(SIO_WAIT_CMD_ACK);
				sioPort->writeAck('N');
				break;
			}
	
		ReleaseMutex(diskMutex);
		}
}

void CSIOProtocol::swapDisks(BYTE diskid1, BYTE diskid2)
{
	bool success = false;
	CDiskImage* di;

	if (diskid1 >= 1 && diskid1 <= SIO_MAX_DISKS && diskid2 >= 1 && diskid2 <= SIO_MAX_DISKS)
	{
		WaitForSingleObject(diskMutex, INFINITE);
		di=disk[diskid1];
		disk[diskid1]=disk[diskid2];
		disk[diskid2]=di;
		innerShell_status("swapped disk %d and %d",diskid1,diskid2);
		ReleaseMutex(diskMutex);
	}
}
