/*
Copyright (c) 2001 Dan Vernon

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; either 
version 2 of the License, or (at your option) any later version. This program is 
distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. You should have received a copy of the GNU
General Public License along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// stateStore.cpp: implementation of the CStateStore class.
//
//////////////////////////////////////////////////////////////////////

#include "StateStore.h"

//////////////////////////////////////////////////////////////////////
// Win32 API thread procs
//////////////////////////////////////////////////////////////////////

DWORD WINAPI queueThreadProc(LPVOID lpParameter)
{
	CStateStore* obj = static_cast<CStateStore*>(lpParameter);
	obj->monitorQueue();
	return 0;
}

DWORD WINAPI sioThreadProc(LPVOID lpParameter)
{
	CSIOProtocol* mysio = static_cast<CSIOProtocol*>(lpParameter);
	mysio->listen();
	return 0;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CStateStore::CStateStore()
{
	LPCSTR src = "CStateStore::CStateStore";

	cancelEvent = CreateEvent(NULL,true,false,NULL);
	errorEvent= CreateEvent(NULL,true,false,NULL);
	debugEvent= CreateEvent(NULL,true,false,NULL);
	traceEvent= CreateEvent(NULL,true,false,NULL);
	logEvent= CreateEvent(NULL,true,false,NULL);
	happyRWModeEvent= CreateEvent(NULL,true,false,NULL);
	queueThreadDeathEvent= CreateEvent(NULL,true,false,NULL);

	speedUpActive=false;

	currentSpeedMutex = CreateMutex(NULL, false, NULL);
	ReleaseMutex(currentSpeedMutex);
	currentSpeed=1;

	controlMutex = CreateMutex(NULL, false, NULL);
	ReleaseMutex(controlMutex);

	sioThread=NULL;
	queueThread=NULL;

	SECURITY_ATTRIBUTES sa;
	sa.bInheritHandle=1;
	sa.lpSecurityDescriptor=0;
	sa.nLength=sizeof(sa);
	DWORD threadID;
	queueThread=(HANDLE)CreateThread(&sa,0,queueThreadProc,
					  				reinterpret_cast<void*>(this),
									0, &threadID );
	if (!queueThread) 
	{
		innerShell_error(GetLastError(),src,"CreateThread",NULL);
	}
	else
	{
		SetThreadPriority(queueThread,THREAD_PRIORITY_IDLE);
	}
}

CStateStore::~CStateStore()
{
	setQueueThreadDeath(true);
	while(!isQueueMonitorStopped()) Sleep(100);
	
	CloseHandle(cancelEvent);
	CloseHandle(errorEvent);
	CloseHandle(debugEvent);
	CloseHandle(traceEvent);
	CloseHandle(logEvent);
	CloseHandle(controlMutex);
	CloseHandle(happyRWModeEvent);

	CloseHandle(queueThreadDeathEvent);

	if (queueThread)
		CloseHandle(queueThread);
	if (sioThread)
		CloseHandle(sioThread);
}

bool CStateStore::isErrorSet()
{
	return isEventSet(errorEvent);
}

bool CStateStore::isQueueThreadDeathSet()
{
	return isEventSet(queueThreadDeathEvent);
}

bool CStateStore::isStopped()
{
	return isSioMonitorStopped();
}

bool CStateStore::isQueueMonitorStopped()
{
	return isEventSet(queueThread);
}

bool CStateStore::isSioMonitorStopped()
{
	return isEventSet(sioThread);
}

bool CStateStore::isStarted()
{
	return !isStopped();
}

bool CStateStore::isCancelSet()
{
	return isEventSet(cancelEvent);
}

void CStateStore::setError(bool flag)
{
	setEvent(errorEvent,flag);
}

void CStateStore::setHappyRWMode(bool flag)
{
	setEvent(happyRWModeEvent,flag);
}
bool CStateStore::isHappyRWModeSet()
{
	return isEventSet(happyRWModeEvent);
}

void CStateStore::setCancel(bool flag)
{
	setEvent(cancelEvent,flag);
}

void CStateStore::setDebug(bool flag)
{
	setEvent(debugEvent,flag);
}

void CStateStore::setTrace(bool flag)
{
	setEvent(traceEvent,flag);
}

void CStateStore::setQueueThreadDeath(bool flag)
{
	setEvent(queueThreadDeathEvent,flag);
}

void CStateStore::setLog(bool flag)
{
	setEvent(logEvent,flag);
}

bool CStateStore::isDebugSet()
{
	return isEventSet(debugEvent);
}

bool CStateStore::isTraceSet()
{
	return isEventSet(traceEvent);
}

bool CStateStore::isLogSet()
{
	return isEventSet(logEvent);
}

void CStateStore::setEvent(HANDLE event, bool flag)
{
	if(flag)
		SetEvent(event);
	else
		ResetEvent(event);
}

bool CStateStore::isEventSet(HANDLE event)
{
	DWORD res;

	if (!event) return true;  // used for thread handles

	res = WaitForSingleObject(event,0);
	switch(res)
	{
    case WAIT_TIMEOUT:
		return false;
        break;

	default: // WAIT_OBJECT_0:
		return true;
		break;
	}
}

void CStateStore::monitorQueue()
{
	innerShell_status("successfully loaded atariSioDriver version %s",VERSION);
	while(!isQueueThreadDeathSet())
	{
		emptyQueue();
		Sleep(100);
	}
}

void CStateStore::emptyQueue()
{
	while(g_evq && g_evq->size()>0)
			g_evq->popEvent();
}

bool CStateStore::stop(bool silent)
{
	if (isStopped()) { if (!silent) innerShell_status("sio emulation is already stopped"); return true; }

	WaitForSingleObject(controlMutex, INFINITE);

	if (queueThread || sioThread)
	{

		setCancel(true);
		while(!isStopped()) Sleep(100);

		CloseHandle(queueThread);
		queueThread=NULL;
		CloseHandle(sioThread);
		sioThread=NULL;
	}

	if (!silent) innerShell_status("sio emulation stopped");

	ReleaseMutex(controlMutex);

	return true;
}

bool CStateStore::start(bool silent)
{
	if (!isStopped()) { if (!silent) innerShell_status("sio emulation is already started"); return true; }

	LPCSTR src = "CStateStore::start";

	stop(true);

	WaitForSingleObject(controlMutex, INFINITE);

	setCancel(false);
	setError(false);

	if (!g_sio->sioPort->portOpened)
	{
		innerShell_error(42,src,"1","cannot start emulation without a valid COM port");
		goto exitPoint;
	}

	SetPriorityClass(GetCurrentProcess(),REALTIME_PRIORITY_CLASS); //REALTIME?
	SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_IDLE);

	SECURITY_ATTRIBUTES sa;
	sa.bInheritHandle=1;
	sa.lpSecurityDescriptor=0;
	sa.nLength=sizeof(sa);
	DWORD threadID;

	sioThread=(HANDLE)CreateThread(&sa,0,sioThreadProc,
					  		   		reinterpret_cast<void*>(g_sio),
									0, &threadID );
	if (!sioThread)
	{
		innerShell_error(GetLastError(),src,"CreateThread",NULL);
		goto exitPoint;
	}

	SetThreadPriority(sioThread,THREAD_PRIORITY_TIME_CRITICAL);
	
	while(isStopped() && !isErrorSet()){Sleep(100);}

	if (!silent && !isStopped()) 
	{
		innerShell_status("sio emulation started on COM%d",g_sio->sioPort->portNumber);
	}

exitPoint:
	ReleaseMutex(controlMutex);

	return !isStopped();
}


void CStateStore::setCurrentSpeed(BYTE newval)
{
	WaitForSingleObject(currentSpeedMutex, INFINITE);
	currentSpeed=newval;
	ReleaseMutex(currentSpeedMutex);
}

BYTE CStateStore::getCurrentSpeed()
{
	BYTE b;
	WaitForSingleObject(currentSpeedMutex, INFINITE);
	b=currentSpeed;
	ReleaseMutex(currentSpeedMutex);
	return b;
}
